from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import polars as pl

from loguru import logger
from pathlib import Path

from .track import Track


class Recording:
    def __init__(
        self,
        recording_id: int,
        recording_meta_file: Path,
        tracks_meta_file: Path,
        tracks_file: Path,
        background_image_file: Path,
        background_image_scale_factor: float = 1.0,
        dataset: str = "unknown-v0.0",
    ) -> None:
        self._recording_id = recording_id
        self._datset_name = dataset
        self._background_image_file = background_image_file
        self._background_image_scale_factor = background_image_scale_factor

        logger.debug("Load csv files for recording {}.", recording_id)

        self._recording_meta_data = pl.read_csv(recording_meta_file).to_dicts()[0]

        self._tracks_meta_data = pl.read_csv(tracks_meta_file)

        self._tracks_file = tracks_file
        self._tracks_data = None

        self._background_image = None

        self._lanelet2_map_file = None
        self._opendrive_map_file = None

        self._track_id_key = "trackId" if "trackId" in self._tracks_meta_data else "id"
        self._is_highd = "trackId" not in self._tracks_meta_data

        self._track_ids = self._tracks_meta_data[self._track_id_key].to_list()

        initial_frame = 0 if not self._is_highd else 1
        try:
            max_frame = int(
                self._recording_meta_data["duration"]
                * self._recording_meta_data["frameRate"]
            )
        except KeyError as e:
            msg = f"{recording_id:02d}_recordingMeta.csv data does not contain key {e}"
            raise KeyError(msg) from e
        self._frames = np.arange(initial_frame, max_frame + 1).tolist()

        self._tracks = {}

    def _read_tracks_file(self, tracks_file: Path) -> pl.DataFrame:
        logger.debug("Load tracks file for recording {}", self._recording_id)

        schema_overrides = None
        if not self._is_highd:
            schema_overrides = {col: pl.Utf8 for col in Track.SEMICOLON_LIST_COLUMNS}
        return pl.read_csv(tracks_file, schema_overrides=schema_overrides)

    @property
    def id(self) -> int:
        return self._recording_id

    @property
    def datset_name(self) -> int:
        return self._datset_name

    @property
    def meta_data_keys(self) -> list[str]:
        return list(self._recording_meta_data)

    @property
    def location_id(self) -> int:
        return self._recording_meta_data["locationId"]

    @property
    def track_ids(self) -> list[int]:
        return self._track_ids

    @property
    def frames(self) -> list[int]:
        return self._frames

    @property
    def lanelet2_map_file(self) -> Path:
        return self._lanelet2_map_file

    @lanelet2_map_file.setter
    def lanelet2_map_file(self, lanelet2_map_file: Path) -> None:
        self._lanelet2_map_file = lanelet2_map_file

    @property
    def opendrive_map_file(self) -> Path:
        return self._opendrive_map_file

    @opendrive_map_file.setter
    def opendrive_map_file(self, opendrive_map_file: Path) -> None:
        self._opendrive_map_file = opendrive_map_file

    def _get_tracks_data(self) -> pl.DataFrame:
        if self._tracks_data is None:
            self._tracks_data = self._read_tracks_file(self._tracks_file)
        return self._tracks_data

    def get_meta_data(self, key: str) -> any:
        if key not in self._recording_meta_data:
            msg = f"Invalid recording meta data key: {key}"
            raise KeyError(msg)

        data = self._recording_meta_data[key]

        return data

    def get_track_ids_at_frame(self, frame: int) -> list[Track]:
        tracks_data = self._get_tracks_data()

        track_ids = (
            tracks_data.filter(pl.col("frame") == frame)
            .select(self._track_id_key)
            .sort(self._track_id_key)
            .to_series()
            .to_list()
        )

        return track_ids

    def get_track(self, track_id: int) -> Track:
        if track_id not in self._track_ids:
            msg = f"Invalid track ID {track_id} for recording {self._recording_id}."
            raise KeyError(msg)

        tracks_data = self._get_tracks_data()

        if track_id not in self._tracks:
            track_meta_data = self._tracks_meta_data.filter(
                pl.col(self._track_id_key) == track_id
            ).to_dicts()[0]
            track_data = tracks_data.filter(pl.col(self._track_id_key) == track_id)
            track = Track(track_id, track_meta_data, track_data)
            self._tracks[track_id] = track

        track = self._tracks[track_id]
        return track

    def get_track_batches(self, track_batch_size: int) -> list[dict]:
        """
        Return a list of track batches (consisting of recording_id, start_track_idx and end_track_idx) for this recording.
        """

        track_batches = []

        n_tracks = len(self._track_ids)

        if track_batch_size == -1:
            n_batches = 1
            track_batch_size = n_tracks
        else:
            n_batches = n_tracks // track_batch_size
            if n_tracks % track_batch_size:
                n_batches += 1

        for i in range(n_batches):
            start_track_idx = i * track_batch_size
            end_track_idx = start_track_idx + track_batch_size - 1

            if end_track_idx > n_tracks:
                end_track_idx = n_tracks - 1

            track_batches.append(
                {
                    "recording_id": self._recording_id,
                    "start_track_idx": start_track_idx,
                    "end_track_idx": end_track_idx,
                }
            )

        return track_batches

    def get_background_image(self) -> np.ndarray:
        if self._background_image is None:
            self._background_image = plt.imread(self._background_image_file)

        return self._background_image

    def plot_track(
        self, track_id: int | list[int], folder: Path, combine: bool = True
    ) -> None:
        if isinstance(track_id, list):
            return self._plot_multiple_tracks(track_id, folder, combine)

        if isinstance(track_id, int):
            return self._plot_single_track(track_id, folder)

        msg = "track_id must be of type int or list[int]"
        raise TypeError(msg)

    def _plot_single_track(self, track_id: int, folder: Path) -> None:
        logger.debug(f"Plot recording {self._recording_id}, track {track_id}")

        track = self.get_track(track_id)
        if self._is_highd:
            background_image_trajectory = track.get_background_image_trajectory(
                0.1, self._background_image_scale_factor
            )
        else:
            background_image_trajectory = track.get_background_image_trajectory(
                self._recording_meta_data["orthoPxToMeter"],
                self._background_image_scale_factor,
            )

        background_image = self.get_background_image()
        plot_file = Path(folder) / f"{self._recording_id}_{track_id}.jpg"

        height, width, _ = background_image.shape
        dpi = 100
        f, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
        ax.imshow(background_image)
        ax.plot(
            background_image_trajectory[:, 0],
            background_image_trajectory[:, 1],
            color="red",
            linewidth=2,
        )
        ax.axis("off")
        f.savefig(plot_file, bbox_inches="tight", pad_inches=0)
        plt.close(f)

        return plot_file

    def _plot_multiple_tracks(
        self, track_ids: list, folder: Path, combine_plots: bool
    ) -> None:
        if combine_plots:
            # Plot all trajectories in one image
            logger.debug(f"Plot recording {self._recording_id}, tracks {track_ids}")

            background_image = self.get_background_image()
            plot_file = (
                Path(folder) / f"{self._recording_id}_{len(track_ids)}_tracks.jpg"
            )

            height, width, _ = background_image.shape
            dpi = 100
            f, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
            ax.imshow(background_image)

            for track_id in track_ids:
                track = self.get_track(track_id)

                if self._is_highd:
                    background_image_trajectory = track.get_background_image_trajectory(
                        0.1, self._background_image_scale_factor
                    )
                else:
                    background_image_trajectory = track.get_background_image_trajectory(
                        self._recording_meta_data["orthoPxToMeter"],
                        self._background_image_scale_factor,
                    )

                    ax.plot(
                        background_image_trajectory[:, 0],
                        background_image_trajectory[:, 1],
                        color="red",
                        linewidth=2,
                    )

            ax.axis("off")
            f.savefig(plot_file, bbox_inches="tight", pad_inches=0)
            plt.close(f)

            return plot_file

        # Plot all trajectories in separate images
        return [self._plot_single_track(track_id, folder) for track_id in track_ids]



from __future__ import annotations

import json
import re
import polars as pl

from loguru import logger
from pathlib import Path

from .recording import Recording


class Dataset:
    TRACKS_FILE_SUFFIX = "_tracks.csv"
    TRACKS_META_FILE_SUFFIX = "_tracksMeta.csv"
    RECORDING_META_FILE_SUFFIX = "_recordingMeta.csv"
    BACKGROUND_IMAGE_FILE_SUFFIX = ("_background.png", "_highway.jpg", "_highway.png")
    POSSIBLE_LANELET_DIR_NAMES = ("lanelet2", "lanelets", "lanelet", "../lanelets")
    POSSIBLE_OPENDRIVE_DIR_NAMES = ("opendrive", ".")

    def __init__(
        self,
        dataset_dir: Path,
        dataset_id: str | None = None,
        major_version_no: int | None = None,
        minor_version_no: int | None = None,
        background_image_scale_factor: float | None = None,
    ) -> None:
        self._dataset_dir = Path(dataset_dir)

        if not self._dataset_dir.exists():
            msg = "Specified dataset dir does not exist."
            raise FileNotFoundError(msg)

        if not self._dataset_dir.is_dir():
            msg = "Specified dataset dir is not a directory."
            raise FileNotFoundError(msg)

        self._dataset_id = dataset_id
        self._major_version_no = major_version_no
        self._minor_version_no = minor_version_no
        if (
            self._dataset_id is None
            or self._major_version_no is None
            or self._minor_version_no is None
        ):
            self._read_dataset_info_from_folder_name()

        logger.debug(
            "Read dataset {}-{} from: {}",
            self._dataset_id,
            self.version,
            self._dataset_dir,
        )

        # Load scale factor
        self._background_image_scale_factor = background_image_scale_factor
        if self._background_image_scale_factor is None:
            self._background_image_scale_factor = (
                self._load_background_image_scale_factor()
            )

        # Data dir
        self._data_dir = self._dataset_dir / "data"
        self._recording_meta_files = {}
        self._tracks_meta_files = {}
        self._tracks_files = {}
        self._background_image_files = {}
        self._recording_ids = []
        self._location_ids = []
        self._recordings_at_location = {}
        self._explore_data_dir()

        # Map dir
        self._maps_dir = self._dataset_dir / "maps"
        self._lanelet2_map_files_per_location = {}
        self._opendrive_map_files_per_location = {}
        self._explore_maps_dir()

    @property
    def id(self) -> str:
        return self._dataset_id

    @property
    def version(self) -> str:
        return f"{self._major_version_no}.{self._minor_version_no}"

    @property
    def recording_ids(self) -> list[int]:
        return self._recording_ids

    @property
    def location_ids(self) -> list[int]:
        return self._location_ids

    @property
    def recordings_at_location(self) -> dict[int, list[int]]:
        return self._recordings_at_location

    def _read_dataset_info_from_folder_name(self) -> None:
        """
        Try to read info from folder name
        Assumption: Format is either
        "<dataset_id>-dataset-v<major_version_nr>.<minor_version_nr>"
        or
        "<dataset_id>-v<major_version_nr>.<minor_version_nr>"
        or
        "<dataset_id>-dataset"
        or
        "<dataset_id>"
        """

        dataset_id = "unknown"
        major_version_no = 0
        minor_version_no = 0

        possible_patterns = {
            "name-dataset-version": "^(.+)-dataset-v([0-9]+).([0-9]+)$",
            "name-version": "^(.+)-v([0-9]+).([0-9]+)$",
            "name-dataset": "^(.+)-dataset$",
        }

        for pattern_name, pattern in possible_patterns.items():  # noqa: B007
            m = re.match(pattern, self._dataset_dir.name)
            if m:
                break

        if m is None:
            dataset_id = self._dataset_dir.name
        else:
            if "name" in pattern_name:
                dataset_id = m.group(1).lower()
            if "version" in pattern_name:
                major_version_no = int(m.group(2))
                minor_version_no = int(m.group(3))

        self._dataset_id = dataset_id
        self._major_version_no = major_version_no
        self._minor_version_no = minor_version_no

    def _load_background_image_scale_factor(self) -> float:
        if self._dataset_id is None:
            logger.warning(
                "Since dataset_id is None, background_image_scale_factor is assumed to be 1.0"
            )
            return 1.0

        scale_factor_json = (
            Path(__file__).parent / "background_image_scale_factors.json"
        )

        with scale_factor_json.open("r") as f:
            scale_factor_dict = json.load(f)

        background_scale_factor = None

        if self._dataset_id in scale_factor_dict:
            version_str = f"{self._major_version_no}.{self._minor_version_no}"
            if version_str in scale_factor_dict[self._dataset_id]:
                background_scale_factor = scale_factor_dict[self._dataset_id][
                    version_str
                ]

        if background_scale_factor is None:
            logger.warning(
                "Dataset id {} could not be found in {}. Scale factor is set to 1.0",
                self._dataset_id,
                scale_factor_json,
            )
            return 1.0

        return background_scale_factor

    def _explore_data_dir(self) -> None:
        logger.debug("Search for recordings in: {}", self._data_dir)

        if not self._data_dir.exists():
            msg = "Data dir does not exist."
            raise FileExistsError(msg)

        if not self._data_dir.is_dir():
            msg = "Data dir is not a directory."
            raise FileExistsError(msg)

        # Find files by recording ids
        def find_files_by_recording_ids(
            data_dir: Path, suffix: str | list[str]
        ) -> dict:
            files_by_recording_ids = {}
            if isinstance(suffix, str):
                suffix = [suffix]
            for sfx in suffix:
                files = data_dir.glob(f"*{sfx}")
                for f in files:
                    recording_id = int(re.match(f"(.*){sfx}", f.name).group(1))
                    files_by_recording_ids[recording_id] = f
            return files_by_recording_ids

        self._recording_meta_files = find_files_by_recording_ids(
            self._data_dir, self.RECORDING_META_FILE_SUFFIX
        )
        self._tracks_meta_files = find_files_by_recording_ids(
            self._data_dir, self.TRACKS_META_FILE_SUFFIX
        )
        self._tracks_files = find_files_by_recording_ids(
            self._data_dir, self.TRACKS_FILE_SUFFIX
        )
        self._background_image_files = find_files_by_recording_ids(
            self._data_dir, self.BACKGROUND_IMAGE_FILE_SUFFIX
        )

        all_potential_recording_ids = sorted(
            set(
                list(self._recording_meta_files.keys())
                + list(self._tracks_meta_files.keys())
                + list(self._tracks_files.keys())
                + list(self._background_image_files.keys())
            )
        )

        # Match files
        self._recording_ids = []
        for recording_id in all_potential_recording_ids:
            if (
                recording_id in self._recording_meta_files
                and recording_id in self._tracks_meta_files
                and recording_id in self._tracks_files
                and recording_id in self._background_image_files
            ):
                self._recording_ids.append(recording_id)

                # Get location id
                recording_meta = pl.read_csv(
                    self._recording_meta_files[recording_id]
                ).to_dicts()[0]
                location_id = recording_meta["locationId"]

                # Store location id
                if location_id not in self._location_ids:
                    self._location_ids.append(location_id)
                    self._recordings_at_location[location_id] = []
                self._recordings_at_location[location_id].append(recording_id)

            else:
                logger.warning("Incomplete data for recording {}", recording_id)

        # Sort location ids
        self._location_ids = sorted(self._location_ids)
        self._recordings_at_location = {
            key: sorted(self._recordings_at_location[key])
            for key in sorted(self._recordings_at_location.keys())
        }

        logger.debug(
            "Found complete data for {} recordings:\n{}",
            len(self._recording_ids),
            self._recording_ids,
        )

    def _explore_maps_dir(self) -> None:
        logger.debug("Search for maps in: {}", self._maps_dir)

        # Lanelet2
        # Find existing lanelet2 dirs
        potential_lanelet2_dirs = []
        for dir_name in self.POSSIBLE_LANELET_DIR_NAMES:
            lanelet2_dir = (self._maps_dir / dir_name).resolve()
            if lanelet2_dir.exists() and lanelet2_dir.is_dir():
                potential_lanelet2_dirs.append(lanelet2_dir)

        if len(potential_lanelet2_dirs) == 0:
            logger.warning("No Lanelet2 map directories found.")
        elif len(potential_lanelet2_dirs) > 1:
            msg = f"Multiple Lanelet2 map directories found: {potential_lanelet2_dirs}. Cannot decide which is the right one."
            raise ValueError(msg)
        else:
            self._find_lanelet2_maps(potential_lanelet2_dirs[0])

        # OpenDRIVE
        # Find existing opendrive dirs
        potential_opendrive_dirs = []
        for dir_name in self.POSSIBLE_OPENDRIVE_DIR_NAMES:
            opendrive_dir = (self._maps_dir / dir_name).resolve()
            if opendrive_dir.exists() and opendrive_dir.is_dir():
                potential_opendrive_dirs.append(opendrive_dir)
                break

        if len(potential_opendrive_dirs) == 0:
            logger.warning("No OpenDrive map directories found.")
        elif len(potential_opendrive_dirs) > 1:
            msg = f"Multiple OpenDrive map directories found: {potential_opendrive_dirs}. Cannot decide which is the right one."
            raise ValueError(msg)
        else:
            self._find_opendrive_maps(potential_opendrive_dirs[0])

    def _find_lanelet2_maps(self, lanelet2_dir: Path) -> None:
        if not lanelet2_dir.exists() or not lanelet2_dir.is_dir():
            logger.warning("Lanelet2 map directory does not exist at {}.", lanelet2_dir)
            return

        potential_lanelet2_maps = list(lanelet2_dir.rglob("*.osm"))

        # Assumption: Format is "X_Y.osm". X=location_id int, Y=location name str
        lanelet2_map_files_per_location = {}
        for f in potential_lanelet2_maps:
            location_id = None

            if f.stem.startswith("location"):
                m = re.match("location.*([0-9]+).osm", f.name)
                if m:
                    location_id = int(m.group(1))
            else:
                m = re.match("([0-9]+).*.osm", f.name)
                if m:
                    location_id = int(m.group(1))

            if location_id is not None:
                lanelet2_map_files_per_location[location_id] = f

        self._lanelet2_map_files_per_location = lanelet2_map_files_per_location

    def _find_opendrive_maps(self, opendrive_dir: Path) -> None:
        if not opendrive_dir.exists() or not opendrive_dir.is_dir():
            logger.warning(
                "OpenDrive map directory does not exist at {}.", opendrive_dir
            )
            return

        potential_opendrive_maps = list(opendrive_dir.rglob("*.xodr"))
        opendrive_map_files_per_location = {}
        for f in potential_opendrive_maps:
            # Assumption:
            # a) Folders with format "X_Y". X=location_id int, Y=location name str.
            # OR
            # b) Folders with format "X". X=location_id int.

            location_id = None

            folder_name = f.parent.name
            if folder_name == opendrive_dir.name:
                continue

            potential_folder_name_patterns = {
                "two_locations": "([0-9]+)_([0-9]+).*",  # special case for exid map
                "one_location": "([0-9]+).*",
            }

            for (
                pattern_name,  # noqa: B007
                folder_name_pattern,
            ) in potential_folder_name_patterns.items():
                m = re.match(folder_name_pattern, folder_name)
                if m:
                    break

            if m:
                if pattern_name == "one_location":
                    location_id = int(m.group(1))
                    opendrive_map_files_per_location[location_id] = f

                elif pattern_name == "two_locations":
                    location_id1 = int(m.group(1))
                    location_id2 = int(m.group(2))

                    opendrive_map_files_per_location[location_id1] = f
                    opendrive_map_files_per_location[location_id2] = f

        self._opendrive_map_files_per_location = opendrive_map_files_per_location

    def get_recording(self, recording_id: str) -> Recording:
        logger.debug("Get recording {}.", recording_id)

        if recording_id not in self._recording_ids:
            msg = (
                f"Invalid recording ID. Available recordings are: {self._recording_ids}"
            )
            raise KeyError(msg)

        recording = Recording(
            recording_id,
            self._recording_meta_files[recording_id],
            self._tracks_meta_files[recording_id],
            self._tracks_files[recording_id],
            self._background_image_files[recording_id],
            self._background_image_scale_factor,
            f"{self.id}-v{self.version}",
        )

        location_id = recording.location_id
        if location_id in self._lanelet2_map_files_per_location:
            lanelet2_map_file = self._lanelet2_map_files_per_location[location_id]
            recording.lanelet2_map_file = lanelet2_map_file
        if location_id in self._opendrive_map_files_per_location:
            opendrive_map_file = self._opendrive_map_files_per_location[location_id]
            recording.opendrive_map_file = opendrive_map_file

        return recording

    def get_track_batches(
        self, track_batch_size: int, recording_ids: list[int] | None = None
    ) -> list[list[int]]:
        """
        Return a list of track batches (consisting of recording_id, start_track_idx and end_track_idx) for all recordings in the dataset.
        """

        track_batches = []

        for recording_id in self._recording_ids:
            if recording_ids is not None and recording_id not in recording_ids:
                continue
            track_batches.extend(
                self.get_recording(recording_id).get_track_batches(track_batch_size)
            )

        return track_batches



import pytest

from lxd_io import Dataset
from lxd_io.track import Track


@pytest.fixture
def valid_track() -> Track:
    return Dataset("test/data/valid-dataset-v1.5").get_recording(1).get_track(1)


@pytest.fixture
def valid_highd_track() -> Track:
    return Dataset("test/data/highD-dataset-v1.0").get_recording(1).get_track(1)


def test_track_access_properties(valid_track: Track, valid_highd_track: Track) -> None:
    assert valid_track.id == 1
    assert len(valid_track.meta_data_keys) > 0
    assert len(valid_track.data_keys) > 0

    assert valid_highd_track.id == 1
    assert len(valid_highd_track.meta_data_keys) > 0
    assert len(valid_highd_track.data_keys) > 0


def test_track_get_meta_data(valid_track: Track, valid_highd_track: Track) -> None:
    # Valid meta_data_keys
    for meta_data_key in valid_track.meta_data_keys:
        assert valid_track.get_meta_data(meta_data_key) is not None
    for meta_data_key in valid_highd_track.meta_data_keys:
        assert valid_highd_track.get_meta_data(meta_data_key) is not None

    # Invalid meta_data_key
    with pytest.raises(KeyError):
        valid_track.get_meta_data("invalid_key")


def test_track_get_data(valid_track: Track, valid_highd_track: Track) -> None:
    # Valid data_keys
    for data_key in valid_track.data_keys:
        assert valid_track.get_data(data_key) is not None
    for data_key in valid_highd_track.data_keys:
        assert valid_highd_track.get_data(data_key) is not None

    # Invalid data_key
    with pytest.raises(KeyError, match="Invalid track data key: invalid_key"):
        valid_track.get_data("invalid_key")


def test_track_get_data_at_frame(valid_track: Track, valid_highd_track: Track) -> None:
    # Valid frames
    for frame in valid_track._frames:
        for data_key in valid_track.data_keys:
            assert valid_track.get_data_at_frame(data_key, frame) is not None
    for frame in valid_highd_track._frames:
        for data_key in valid_highd_track.data_keys:
            assert valid_highd_track.get_data_at_frame(data_key, frame) is not None

    # Invalid data_key
    with pytest.raises(KeyError, match="Invalid track data key: invalid_key"):
        valid_track.get_data_at_frame("invalid_key", 0)

    # Invalid frame
    with pytest.raises(KeyError, match=r"Invalid frame: -1"):
        valid_track.get_data_at_frame("xCenter", -1)


def test_track_get_data_between_frames(
    valid_track: Track, valid_highd_track: Track
) -> None:
    # Valid frames
    for frame in valid_track._frames:
        for data_key in valid_track.data_keys:
            assert (
                valid_track.get_data_between_frames(data_key, frame, frame) is not None
            )
    for frame in valid_highd_track._frames:
        for data_key in valid_highd_track.data_keys:
            assert (
                valid_highd_track.get_data_between_frames(data_key, frame, frame)
                is not None
            )

    # Invalid data_key
    with pytest.raises(KeyError):
        valid_track.get_data_between_frames("invalid_key", 0, 0)

    # Invalid frame
    with pytest.raises(KeyError):
        valid_track.get_data_between_frames("xCenter", -1, -1)

    # Invalid frame
    with pytest.raises(KeyError):
        valid_track.get_data_between_frames("xCenter", 0, -1)

    # Invalid frame
    with pytest.raises(KeyError):
        valid_track.get_data_between_frames("xCenter", -1, 0)

    # Invalid frame
    with pytest.raises(KeyError):
        valid_track.get_data_between_frames("xCenter", 1, 0)


def test_track_get_local_trajectory(
    valid_track: Track, valid_highd_track: Track
) -> None:
    assert valid_track.get_local_trajectory() is not None
    assert valid_highd_track.get_local_trajectory() is not None


def test_track_get_utm_image_trajectory(
    valid_track: Track, valid_highd_track: Track
) -> None:
    assert valid_track.get_utm_image_trajectory(0, 0) is not None
    assert valid_highd_track.get_utm_image_trajectory(0, 0) is not None


def test_track_get_bbox_at_frame(valid_track: Track, valid_highd_track: Track) -> None:
    # Valid frames
    for frame in valid_track._frames:
        assert valid_track.get_bbox_at_frame(frame) is not None
        assert valid_track.get_bbox_at_frame(frame, visualization_mode=True) is not None
    for frame in valid_highd_track._frames:
        assert valid_highd_track.get_bbox_at_frame(frame) is not None
        assert (
            valid_highd_track.get_bbox_at_frame(frame, visualization_mode=True)
            is not None
        )

    # Invalid frame
    with pytest.raises(KeyError):
        valid_track.get_bbox_at_frame(-1)



import numpy as np
import pytest

from pathlib import Path

from lxd_io.recording import Recording


@pytest.fixture
def valid_recording() -> Recording:
    return Recording(
        1,
        "test/data/valid-dataset-v1.5/data/01_recordingMeta.csv",
        "test/data/valid-dataset-v1.5/data/01_tracksMeta.csv",
        "test/data/valid-dataset-v1.5/data/01_tracks.csv",
        "test/data/valid-dataset-v1.5/data/01_background.png",
        4.0,
        "valid",
    )


@pytest.fixture
def valid_highd_recording() -> Recording:
    return Recording(
        1,
        "test/data/highD-dataset-v1.0/data/01_recordingMeta.csv",
        "test/data/highD-dataset-v1.0/data/01_tracksMeta.csv",
        "test/data/highD-dataset-v1.0/data/01_tracks.csv",
        "test/data/highD-dataset-v1.0/data/01_highway.png",
        4.0,
        "highd",
    )


@pytest.fixture
def lanelet2_path() -> Path:
    return Path("test/data/valid-dataset-v1.5/maps/lanelet2/1_location.osm")


@pytest.fixture
def opendrive_path() -> Path:
    return Path("test/data/valid-dataset-v1.5/maps/opendrive/1/1_location.xodr")


def test_recording_invalid_recording_meta_csv() -> None:
    with pytest.raises(KeyError):
        Recording(
            1,
            "test/data/invalid-dataset/data/1_recordingMeta.csv",
            "test/data/invalid-dataset/data/1_tracksMeta.csv",
            "test/data/invalid-dataset/data/1_tracks.csv",
            "test/data/invalid-dataset/data/1_background.png",
            0.0,
            "invalid",
        )


def test_recording_access_properties(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    assert valid_recording.id == 1
    assert valid_recording.datset_name == "valid"
    assert len(valid_recording.meta_data_keys) > 0
    assert valid_recording.location_id == 1
    assert len(valid_recording.track_ids) > 0
    assert len(valid_recording.frames) > 0
    assert valid_recording.lanelet2_map_file is None
    assert valid_recording.opendrive_map_file is None

    assert valid_highd_recording.id == 1
    assert valid_highd_recording.datset_name == "highd"
    assert len(valid_highd_recording.meta_data_keys) > 0
    assert valid_highd_recording.location_id == 2
    assert len(valid_highd_recording.track_ids) > 0
    assert len(valid_highd_recording.frames) > 0
    assert valid_highd_recording.lanelet2_map_file is None
    assert valid_highd_recording.opendrive_map_file is None


def test_recording_set_map_paths(
    valid_recording: Recording, lanelet2_path: Path, opendrive_path: Path
) -> None:
    valid_recording.lanelet2_map_file = lanelet2_path
    assert valid_recording.lanelet2_map_file == lanelet2_path

    valid_recording.opendrive_map_file = opendrive_path
    assert valid_recording.opendrive_map_file == opendrive_path


def test_recording_get_meta_data(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    # Valid meta_data_keys
    for meta_data_key in valid_recording.meta_data_keys:
        assert valid_recording.get_meta_data(meta_data_key) is not None
    for meta_data_key in valid_highd_recording.meta_data_keys:
        assert valid_highd_recording.get_meta_data(meta_data_key) is not None

    # Invalid meta_data_key
    with pytest.raises(KeyError):
        valid_recording.get_meta_data("invalid_key")


def test_recording_get_track_ids_at_frame(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    # Valid frames
    for frame in valid_recording.frames:
        assert len(valid_recording.get_track_ids_at_frame(frame)) > 0
    for frame in valid_highd_recording.frames:
        assert len(valid_highd_recording.get_track_ids_at_frame(frame)) > 0

    # Invalid frame
    assert len(valid_recording.get_track_ids_at_frame(-1)) == 0


def test_recording_get_track(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    # Valid track_ids
    for track_id in valid_recording.track_ids:
        assert valid_recording.get_track(track_id) is not None
    for track_id in valid_highd_recording.track_ids:
        assert valid_highd_recording.get_track(track_id) is not None

    # Invalid track_id
    with pytest.raises(KeyError):
        valid_recording.get_track(-1)


def test_recording_get_track_batches(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    # Valid track_batch_size
    # Batches with 1 track
    assert len(valid_recording.get_track_batches(1)) == 2
    # 1 batch of all tracks
    assert len(valid_recording.get_track_batches(-1)) == 1

    # highd
    # Batches with 1 track
    assert len(valid_highd_recording.get_track_batches(1)) == 2
    # 1 batch of all tracks
    assert len(valid_highd_recording.get_track_batches(-1)) == 1

    # Invalid track_batch_size
    assert len(valid_recording.get_track_batches(-2)) == 0


def test_recording_get_background_image(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    assert isinstance(valid_recording.get_background_image(), np.ndarray)
    assert isinstance(valid_highd_recording.get_background_image(), np.ndarray)


def test_recording_plot_track(
    valid_recording: Recording, valid_highd_recording: Recording
) -> None:
    # One track_id
    valid_recording.plot_track(1, "test")
    # Multiple track_ids (combined image)
    valid_recording.plot_track([1, 2], "test")
    # Multiple track_ids (separate images)
    valid_recording.plot_track([1, 2], "test", combine=False)

    # highd
    # One track_id
    valid_highd_recording.plot_track(1, "test")
    # Multiple track_ids (combined image)
    valid_highd_recording.plot_track([1, 2], "test")
    # Multiple track_ids (separate images)
    valid_highd_recording.plot_track([1, 2], "test", combine=False)

    # Invalid track_id
    with pytest.raises(KeyError):
        valid_recording.plot_track(-1, "test")
    with pytest.raises(TypeError):
        valid_recording.plot_track("1", "test")
    with pytest.raises(KeyError):
        valid_recording.plot_track([-1, -2], "test")
    with pytest.raises(KeyError):
        valid_recording.plot_track([-1, -2], "test", combine=False)



from __future__ import annotations

import numpy as np
import polars as pl

from typing import TypeVar

T = TypeVar("T", int, float)


class Track:
    MAX_CONFLICTING_LANELETS = 5

    SEMICOLON_LIST_COLUMNS = (
        "leftAlongsideId",
        "rightAlongsideId",
        "laneletId",
        "latLaneCenterOffset",
        "lonLaneletPos",
        "laneletLength",
        "laneWidth",
    )

    _SEMICOLON_LIST_COLUMN_DTYPE = (
        ("leftAlongsideId", int),
        ("rightAlongsideId", int),
        ("laneletId", int),
        ("latLaneCenterOffset", float),
        ("lonLaneletPos", float),
        ("laneletLength", float),
        ("laneWidth", float),
    )

    SEMICOLON_LIST_COLUMN_DEFAULT = (
        ("leftAlongsideId", -1),
        ("rightAlongsideId", -1),
        ("laneletId", -1),
        ("latLaneCenterOffset", np.nan),
        ("lonLaneletPos", np.nan),
        ("laneletLength", np.nan),
        ("laneWidth", np.nan),
    )

    def __init__(
        self, track_id: int, track_meta_data: dict, track_data: pl.DataFrame
    ) -> None:
        self._track_id = track_id
        self._track_meta_data = track_meta_data
        self._track_data = track_data

        if "xCenter" in self._track_data:
            self._is_highd = False
        else:
            self._is_highd = True

        self._frames = self._track_data["frame"].to_list()
        self._semicolon_list_data = {}

    @property
    def id(self) -> int:
        return self._track_id

    @property
    def meta_data_keys(self) -> list:
        return list(self._track_meta_data.keys())

    @property
    def data_keys(self) -> list:
        return list(self._track_data.columns)

    def get_meta_data(self, key: str) -> any:
        if key not in self._track_meta_data:
            msg = f"Invalid track meta data key: {key}"
            raise KeyError(msg)

        data = self._track_meta_data[key]

        return data

    def _get_semicolon_list_df(self, key: str) -> pl.DataFrame:
        """
        Helper method to process the semicolon list columns on demand.
        """

        def semicolon_list_str_to_fixed_len_list(
            semicolon_list_str: str,
            fixed_len: int = self.MAX_CONFLICTING_LANELETS,
            dtype: type[T] = float,
            fill_value: T = np.nan,
        ) -> list[float]:
            output_list = [fill_value] * fixed_len
            if semicolon_list_str:
                if ";" in semicolon_list_str:
                    for i, v in enumerate(semicolon_list_str.split(";")):
                        output_list[i] = dtype(v)
                else:
                    output_list[0] = dtype(semicolon_list_str)
            return output_list

        if key not in self._semicolon_list_data:
            dtype = dict(self._SEMICOLON_LIST_COLUMN_DTYPE)[key]
            self._semicolon_list_data[key] = self._track_data.select(
                pl.col("frame"), pl.col(key)
            ).with_columns(
                pl.col(key).map_elements(
                    lambda semicolon_list_str: semicolon_list_str_to_fixed_len_list(
                        semicolon_list_str,
                        dtype=dtype,
                        fill_value=dict(self.SEMICOLON_LIST_COLUMN_DEFAULT)[key],
                    ),
                    return_dtype=pl.List(pl.Int64 if dtype is int else pl.Float64),
                )
            )

        return self._semicolon_list_data[key]

    def get_data(self, key: str) -> np.ndarray:
        if key not in self._track_data.columns:
            msg = f"Invalid track data key: {key}"
            raise KeyError(msg)

        if key in self.SEMICOLON_LIST_COLUMNS:
            data = np.array(self._get_semicolon_list_df(key).to_series().to_list())
        else:
            data = self._track_data[key].to_numpy()

        return data

    def get_data_at_frame(self, key: str, frame: int) -> any:
        if key not in self._track_data.columns:
            msg = f"Invalid track data key: {key}"
            raise KeyError(msg)

        if not (self._frames[0] <= frame <= self._frames[-1]):
            msg = f"Invalid frame: {frame}"
            raise KeyError(msg)

        if key in self.SEMICOLON_LIST_COLUMNS:
            data = (
                self._get_semicolon_list_df(key)
                .filter(pl.col("frame") == frame)[key][0]
                .to_list()
            )
        else:
            data = self._track_data.filter(pl.col("frame") == frame)[key][0]

        return data

    def get_data_between_frames(self, key: str, frame0: int, frame1: int) -> any:
        if key not in self._track_data.columns:
            msg = f"Invalid track data key: {key}"
            raise KeyError(msg)

        for frame in (frame0, frame1):
            if not (self._frames[0] <= frame <= self._frames[-1]):
                msg = f"Invalid frame: {frame}"
                raise KeyError(msg)

        if frame0 > frame1:
            msg = f"Invalid frame range: {frame0} > {frame1}"
            raise KeyError(msg)

        if key in self.SEMICOLON_LIST_COLUMNS:
            data = np.array(
                self._get_semicolon_list_df(key)
                .filter(pl.col("frame").is_between(frame0, frame1))[key]
                .to_list()
            )
        else:
            data = (
                self._track_data.filter(pl.col("frame").is_between(frame0, frame1))
                .sort("frame")[key]
                .to_numpy()
            )

        return data

    def get_local_trajectory(self) -> np.ndarray:
        if self._is_highd:
            x = self._track_data["x"] + self._track_data["width"] / 2
            y = self._track_data["y"] + self._track_data["height"] / 2
        else:
            x = self._track_data["xCenter"]
            y = self._track_data["yCenter"]

        trajectory = np.vstack((x, y)).T

        return trajectory

    def get_utm_image_trajectory(
        self, utm_origin_x: float, utm_origin_y: float
    ) -> np.ndarray:
        if self._is_highd:
            return self.get_local_trajectory()

        x_utm = self._track_data["xCenter"] + utm_origin_x
        y_utm = self._track_data["yCenter"] + utm_origin_y

        utm_trajectory = np.vstack((x_utm, y_utm)).T

        return utm_trajectory

    def get_background_image_trajectory(
        self, px2m_factor: float, background_image_scale_factor: float = 1.0
    ) -> np.ndarray:
        if self._is_highd:
            x_image = (
                self._track_data["x"] + self._track_data["width"] / 2
            ) / px2m_factor
            y_image = (
                self._track_data["y"] + self._track_data["height"] / 2
            ) / px2m_factor
        else:
            x_image = self._track_data["xCenter"] / px2m_factor
            y_image = -self._track_data["yCenter"] / px2m_factor
        image_trajectory = np.vstack((x_image, y_image)).T

        # Scale down factor
        image_trajectory = image_trajectory / background_image_scale_factor

        return image_trajectory

    def get_bbox_at_frame(
        self, frame: int, visualization_mode: bool = False, ortho_px_to_m: float = 0.1
    ) -> np.ndarray:
        if self._is_highd:
            length = self.get_meta_data("width")
            width = self.get_meta_data("height")
            x_center = self.get_data_at_frame("x", frame) + length / 2
            y_center = self.get_data_at_frame("y", frame) + width / 2
            heading = (self.get_meta_data("drivingDirection") - 2) * 180

        else:
            x_center = self.get_data_at_frame("xCenter", frame)
            y_center = self.get_data_at_frame("yCenter", frame)
            length = self.get_data_at_frame("length", frame)
            width = self.get_data_at_frame("width", frame)
            heading = self.get_data_at_frame("heading", frame)

        if visualization_mode:
            x_center = x_center / ortho_px_to_m
            y_center = -y_center / ortho_px_to_m
            length = length / ortho_px_to_m
            width = width / ortho_px_to_m
            heading = heading * (-1)
            if heading < 0:
                heading += 360

        return self._bbox_to_vertices(
            x_center, y_center, length, width, np.deg2rad(heading)
        )

    @staticmethod
    def _bbox_to_vertices(
        x_center: np.ndarray,
        y_center: np.ndarray,
        length: np.ndarray,
        width: np.ndarray,
        heading: np.ndarray,
    ) -> np.ndarray:
        """
        Calculate the corners of a rotated bbox from the position, shape and heading for every timestamp.

        :param x_center: x coordinates of the object center positions [num_timesteps]
        :param y_center: y coordinates of the object center positions [num_timesteps]
        :param length: objects lengths [num_timesteps]
        :param width: object widths [num_timesteps]
        :param heading: object heading (rad) [num_timesteps]
        :return: Numpy array in the shape [num_timesteps, 4 (corners), 2 (dimensions)]
        """
        centroids = np.column_stack([x_center, y_center])

        # Precalculate all components needed for the corner calculation
        half_length = length / 2
        half_width = width / 2
        heading_cos = np.cos(heading)
        heading_sin = np.sin(heading)

        lc = half_length * heading_cos
        ls = half_length * heading_sin
        wc = half_width * heading_cos
        ws = half_width * heading_sin

        # Calculate all four rotated bbox corner positions assuming the object is located at the origin.
        # To do so, rotate the corners at [+/- length/2, +/- width/2] as given by the orientation.
        # Use a vectorized approach using precalculated components for maximum efficiency
        rotated_bbox_vertices = np.empty((centroids.shape[0], 4, 2))

        # Front-right corner
        rotated_bbox_vertices[:, 0, 0] = lc - ws
        rotated_bbox_vertices[:, 0, 1] = ls + wc

        # Rear-right corner
        rotated_bbox_vertices[:, 1, 0] = -lc - ws
        rotated_bbox_vertices[:, 1, 1] = -ls + wc

        # Rear-left corner
        rotated_bbox_vertices[:, 2, 0] = -lc + ws
        rotated_bbox_vertices[:, 2, 1] = -ls - wc

        # Front-left corner
        rotated_bbox_vertices[:, 3, 0] = lc + ws
        rotated_bbox_vertices[:, 3, 1] = ls - wc

        # Move corners of rotated bounding box from the origin to the object's location
        rotated_bbox_vertices = rotated_bbox_vertices + np.expand_dims(
            centroids, axis=1
        )

        return rotated_bbox_vertices



import pytest
from pathlib import Path
from lxd_io import Dataset
from lxd_io.recording import Recording


@pytest.fixture
def invalid_dataset_dir() -> Path:
    return Path("test/data/invalid-dataset")


@pytest.fixture
def invalid_dataset_recording_meta_csv() -> Path:
    return Path("test/data/invalid-dataset/data/1_recordingMeta.csv")


@pytest.fixture
def valid_dataset_with_version_dir() -> Path:
    return Path("test/data/valid-dataset-v1.5")


@pytest.fixture
def valid_dataset_without_version_dir() -> Path:
    return Path("test/data/valid-dataset")


@pytest.fixture
def valid_dataset_only_name_dir() -> Path:
    return Path("test/data/valid_dataset")


@pytest.fixture
def valid_highd_like_dataset_dir() -> Path:
    return Path("test/data/highD-dataset-v1.0")


@pytest.fixture
def ind_1_0_like_dataset_dir() -> Path:
    return Path("test/data/inD-dataset-v1.0")


@pytest.fixture
def ind_1_1_like_dataset_dir() -> Path:
    return Path("test/data/inD-dataset-v1.1")


@pytest.fixture
def round_1_0_like_dataset_dir() -> Path:
    return Path("test/data/rounD-dataset-v1.0")


@pytest.fixture
def round_1_1_like_dataset_dir() -> Path:
    return Path("test/data/rounD-dataset-v1.1")


@pytest.fixture
def unid_1_0_like_dataset_dir() -> Path:
    return Path("test/data/uniD-dataset-v1.0")


@pytest.fixture
def unid_1_1_like_dataset_dir() -> Path:
    return Path("test/data/uniD-dataset-v1.1")


@pytest.fixture
def exid_1_0_like_dataset_dir() -> Path:
    return Path("test/data/exiD-dataset-v1.0")


@pytest.fixture
def exid_1_1_like_dataset_dir() -> Path:
    return Path("test/data/exiD-dataset-v1.1")


@pytest.fixture
def exid_1_2_like_dataset_dir() -> Path:
    return Path("test/data/exiD-dataset-v1.2")


@pytest.fixture
def exid_2_0_like_dataset_dir() -> Path:
    return Path("test/data/exiD-dataset-v2.0")


@pytest.fixture
def exid_2_1_like_dataset_dir() -> Path:
    return Path("test/data/exiD-dataset-v2.1")


def test_dataset_initialization(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    invalid_dataset_recording_meta_csv: Path,
    valid_highd_like_dataset_dir: Path,
) -> None:
    for dataset_dir in (
        invalid_dataset_dir,
        valid_dataset_with_version_dir,
        valid_dataset_without_version_dir,
        valid_dataset_only_name_dir,
        valid_highd_like_dataset_dir,
    ):
        Dataset(dataset_dir)

    with pytest.raises(FileNotFoundError):
        Dataset(Path("/path/to/invalid/dataset"))

    with pytest.raises(FileNotFoundError):
        Dataset(invalid_dataset_recording_meta_csv)


def test_dataset_read_dataset_info_from_folder_name(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    valid_highd_like_dataset_dir: Path,
) -> None:
    dataset = Dataset(invalid_dataset_dir)
    assert dataset.id == "invalid"
    assert dataset.version == "0.0"

    dataset = Dataset(valid_dataset_with_version_dir)
    assert dataset.id == "valid"
    assert dataset.version == "1.5"

    dataset = Dataset(valid_dataset_without_version_dir)
    assert dataset.id == "valid"
    assert dataset.version == "0.0"

    dataset = Dataset(valid_dataset_only_name_dir)
    assert dataset.id == "valid_dataset"
    assert dataset.version == "0.0"

    dataset = Dataset(valid_highd_like_dataset_dir)
    assert dataset.id == "highd"
    assert dataset.version == "1.0"


def test_dataset_load_background_image_scale_factor(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    valid_highd_like_dataset_dir: Path,
) -> None:
    for dataset_dir in (
        invalid_dataset_dir,
        valid_dataset_with_version_dir,
        valid_dataset_without_version_dir,
        valid_dataset_only_name_dir,
    ):
        dataset = Dataset(dataset_dir)
        assert dataset._background_image_scale_factor == 1.0

    dataset = Dataset(valid_highd_like_dataset_dir)
    assert dataset._background_image_scale_factor == 4.0


def test_dataset_explore_data_dir(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    valid_highd_like_dataset_dir: Path,
) -> None:
    dataset = Dataset(invalid_dataset_dir)
    assert len(dataset.recording_ids) == 1
    assert len(dataset.location_ids) == 1
    assert len(dataset.recordings_at_location) == 1

    for dataset_dir in (
        valid_dataset_with_version_dir,
        valid_dataset_without_version_dir,
        valid_dataset_only_name_dir,
    ):
        dataset = Dataset(dataset_dir)
        assert len(dataset.recording_ids) == 2
        assert len(dataset.location_ids) == 1
        assert len(dataset.recordings_at_location) == 1

    dataset = Dataset(valid_highd_like_dataset_dir)
    assert len(dataset.recording_ids) == 1
    assert len(dataset.location_ids) == 1
    assert len(dataset.recordings_at_location) == 1


def test_dataset_explore_maps_dir(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    valid_highd_like_dataset_dir: Path,
    ind_1_0_like_dataset_dir: Path,
    ind_1_1_like_dataset_dir: Path,
    round_1_0_like_dataset_dir: Path,
    round_1_1_like_dataset_dir: Path,
    unid_1_0_like_dataset_dir: Path,
    unid_1_1_like_dataset_dir: Path,
    exid_1_0_like_dataset_dir: Path,
    exid_1_1_like_dataset_dir: Path,
    exid_1_2_like_dataset_dir: Path,
    exid_2_0_like_dataset_dir: Path,
    exid_2_1_like_dataset_dir: Path,
) -> None:
    dataset = Dataset(invalid_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 1
    assert len(dataset._opendrive_map_files_per_location) == 0

    for dataset_dir in (
        valid_dataset_with_version_dir,
        valid_dataset_without_version_dir,
        valid_dataset_only_name_dir,
    ):
        dataset = Dataset(dataset_dir)
        assert len(dataset._lanelet2_map_files_per_location) == 1
        assert len(dataset._opendrive_map_files_per_location) == 1

    # Test all known dataset structures of the public datasets

    # highd
    dataset = Dataset(valid_highd_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 0
    assert len(dataset._opendrive_map_files_per_location) == 0

    # ind 1.0
    dataset = Dataset(ind_1_0_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 2
    assert len(dataset._opendrive_map_files_per_location) == 0

    # ind 1.1
    dataset = Dataset(ind_1_1_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 2
    assert len(dataset._opendrive_map_files_per_location) == 2

    # round 1.0
    dataset = Dataset(round_1_0_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 1
    assert len(dataset._opendrive_map_files_per_location) == 0

    # round 1.1
    dataset = Dataset(round_1_1_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 2
    assert len(dataset._opendrive_map_files_per_location) == 2

    # unid 1.0
    dataset = Dataset(unid_1_0_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 0
    assert len(dataset._opendrive_map_files_per_location) == 0

    # unid 1.1
    dataset = Dataset(unid_1_1_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 1
    assert len(dataset._opendrive_map_files_per_location) == 1

    # exid 1.0
    dataset = Dataset(exid_1_0_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 0
    assert len(dataset._opendrive_map_files_per_location) == 3

    # exid 1.1
    dataset = Dataset(exid_1_1_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 0
    assert len(dataset._opendrive_map_files_per_location) == 3

    # exid 1.2
    dataset = Dataset(exid_1_2_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 0
    assert len(dataset._opendrive_map_files_per_location) == 3

    # exid 2.0
    dataset = Dataset(exid_2_0_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 3
    assert len(dataset._opendrive_map_files_per_location) == 3

    # exid 2.1
    dataset = Dataset(exid_2_1_like_dataset_dir)
    assert len(dataset._lanelet2_map_files_per_location) == 3
    assert len(dataset._opendrive_map_files_per_location) == 3


def test_dataset_get_recording(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    valid_highd_like_dataset_dir: Path,
) -> None:
    # Invalid recording data
    dataset = Dataset(invalid_dataset_dir)
    for recording_id in dataset.recording_ids:
        with pytest.raises(KeyError):
            dataset.get_recording(recording_id)
    # Invalid recording id
    with pytest.raises(KeyError):
        dataset.get_recording(10)

    for dataset_dir in (
        valid_dataset_with_version_dir,
        valid_dataset_without_version_dir,
        valid_dataset_only_name_dir,
        valid_highd_like_dataset_dir,
    ):
        dataset = Dataset(dataset_dir)
        for recording_id in dataset.recording_ids:
            assert isinstance(dataset.get_recording(recording_id), Recording)


def test_dataset_get_track_batches(
    invalid_dataset_dir: Path,
    valid_dataset_with_version_dir: Path,
    valid_dataset_without_version_dir: Path,
    valid_dataset_only_name_dir: Path,
    valid_highd_like_dataset_dir: Path,
) -> None:
    dataset = Dataset(invalid_dataset_dir)
    with pytest.raises(KeyError):
        dataset.get_track_batches(10)
    with pytest.raises(KeyError):
        dataset.get_track_batches(10, [1])

    for dataset_dir in (
        valid_dataset_with_version_dir,
        valid_dataset_without_version_dir,
        valid_dataset_only_name_dir,
    ):
        dataset = Dataset(dataset_dir)
        assert len(dataset.get_track_batches(1)) == 3
        assert len(dataset.get_track_batches(1, [1])) == 2
        # Non-existing recording_id
        assert len(dataset.get_track_batches(1, [1, 5])) == len(
            dataset.get_track_batches(1, [1])
        )

    dataset = Dataset(valid_highd_like_dataset_dir)
    assert len(dataset.get_track_batches(1)) == 2
    assert len(dataset.get_track_batches(1, [1])) == 2
    # Non-existing recording_id
    assert len(dataset.get_track_batches(1, [1, 5])) == len(
        dataset.get_track_batches(1, [1])
    )
